from autosync import lib
from click import echo
from meraki.exceptions import AsyncAPIError

async def reSYNC(_db: object, _net_id: str,network: object,task: str, product: str, _appcfg: object):
    """
         Function reSyces a spefic meraki dashbaord eliment
    Args:
        _db(object): Meraki dashboard SDK Objeect
        _net_id(str): Network Id String
        network(object): Data object for current network
        task(str): Task to resync
        product(str): Meraki Product being synced I.E. switch, wirless
        _appcfg(object): Application Config Object
    Returns:
       Nothing Updateds the network Object that is passed
    """
    try:
        action = getattr(_db, f'get{task}')
        setattr(network, task, await action(_net_id))
    except AsyncAPIError as apie:
        echo(
            f'\t {lib.bc.FAIL} Error Running Setting {task} '
            f'{lib.bc.WARNING}Error Message: {str(apie)}{lib.bc.Default}')
    except Exception as e:
        echo(
                f'{lib.bc.WARNING}Error with Module: {str(e)}'
                f'{lib.bc.Default}')
                


from datetime import datetime
import pickle
from os import path, remove
from click import echo
def cacheAge(self):
	return datetime.utcnow() - self.last_sync


# returns the cached object if exists otherwise returns the locally synced
def loadCache(org,appcfg):
	f = f'{appcfg.CACHE_DIR}/{str(org.id)}-{str(org.name)}.mnet'
	if path.exists(f) and appcfg.USE_CACHE:
		with open (f, "rb") as pFile:
			org_cached = pickle.load(pFile)
		if org_cached.lastsync is None:
			echo(f'Has Cache! But it is stale, re-syncing')
			clearCache(org,appcfg)
			org.cached = False

		elif (datetime.utcnow() - org_cached.lastsync) < appcfg.checkCache():
			org.__dict__ = org_cached.__dict__.copy()

		else:
			print(f'Cache File Error! But it is stale, re-syncing')
			clearCache(appcfg)
			org.cached = False

	else:
		echo('No Cahce Found')
		org.cached = False

		
		
# writes it to disk for faster load times
def storeCache(org,appcfg):
	if not appcfg.USE_CACHE: return
	org.cached = True
	f = f'{appcfg.CACHE_DIR}/{str(org.id)}-{str(org.name)}.mnet'
	pickle.dump(org, open(f, "wb"))


# clears cache and kills disk backup
def clearCache(org,appcfg):
	f = f'{appcfg.CACHE_DIR}{str(org.id)}-{str(org.name)} .mnet'
	if path.exists(f):
		remove(f)

# master SYNC function



import asyncio
import datetime
import os
import time
from autosync import model
from autosync import lib
from autosync import utils
from tabulate import tabulate
import pandas as pd
from dateutil import tz
from click import echo

async def RUN(configFile=None,override=False):
    start_time = time.perf_counter()
    echo(f'Start time: {start_time:0.5f}')
    mnets = {}
    masternet = {}
    cfg = model.APPCONFIG(configFile,override)
    mpi = lib.MerakiApi(cfg)
    db = mpi.getAPI()
    echo("LOADING CONFIG FROM ENV")
    masterOrg = model.ORGDB(cfg.tag_master, cfg.tag_target)
    masternet.update({'master': masterOrg})
    masternet['master'].networks.update({'master': 'Master'})
    if cfg.USE_CACHE:
        lib.loadCache(masterOrg, cfg)
    if cfg.ALL_ORGS:
        await utils.getOrginizationsAll(db, mnets)
    else:
        orgList = cfg.whitelist
        orgdbTasks = [
            utils.getOrginizationsWhiteList(db, org, mnets) for org in orgList
        ]
        await asyncio.gather(*orgdbTasks)

    orgThreads = []
    validateThread = []
    for org in mnets:
        orgThreads.append(
            utils.OrgSyncProcessor(mnets[org], cfg, masternet['master'], db))
    [orgThread.start() for orgThread in orgThreads]
    [orgThread.join() for orgThread in orgThreads]
    for org in mnets:
        validateThread.append(
            utils.ValidateOrginization(mnets[org], cfg, masternet['master']))
    [thread.start() for thread in validateThread]
    [thread.join() for thread in validateThread]

    if cfg.USE_CACHE:
        masterOrg.lastsync = datetime.datetime.utcnow()
        lib.storeCache(masterOrg, cfg)
    #heckTasks = [mnetutils]

# await utils.batchSync(mnets,masternet,cfg,stdb,db)
    elapsed_time = time.perf_counter() - start_time
    echo(f'Total Job Runtime: {elapsed_time:0.5f} secounds')
    orgcount = 0
    networkCount = 0
    t = []
    tz_local = tz.tzlocal()
    for org in mnets:
        orgcount = orgcount + 1
        networkCount = networkCount + int(len(mnets[org].networks))
        t.append({
            'Orginization Name': mnets[org].name,
            'Total Network': len(mnets[org].networks),
            'Sync Runtime': mnets[org].syncruntime,
            'Last Sync': mnets[org].lastsync
        })
    echo(
        f'Total Orgs Synced: {orgcount} Total Network Synced: {networkCount}')
    table = pd.DataFrame.from_dict(t)
    echo(tabulate(table, headers='keys', tablefmt='psql'))

if __name__ == '__main__':
    asyncio.run(RUN())
    echo('Done')






async def idFromName (listDicts, name):
	for ld in listDicts:
		if ld['name'] == name:
			return ld  # ld['groupPolicyId']
	return None


# returns object in list where "name" matches <name>
async def matchGidByName(listDicts, gpid):
	for ld in listDicts:
		if 'groupPolicyId' in ld and ld['groupPolicyId'] == gpid:
			return ld
	return None


import meraki.aio
import logging
import meraki
import os
import asyncio
class MerakiAsyncApi:
    def __init__(self,appcfg):
        self._appcfg = appcfg
    def __enter__(self):
        self.api = meraki.aio.AsyncDashboardAPI(
            api_key=self._appcfg.MERAKI_DASHBOARD_API_KEY,
            base_url=self._appcfg.meraki_base_url,
            simulate=self._appcfg.simulate,
            wait_on_rate_limit=self._appcfg.wait_on_rate_limit,
            maximum_concurrent_requests =self._appcfg.maximum_concurrent_requests,
            nginx_429_retry_wait_time = self._appcfg.nginx_429_retry_wait_time,
            maximum_retries=int(self._appcfg.maximum_retries),
            log_file_prefix=os.path.basename(__file__)[:-3],
            log_path="Logs/",
            suppress_logging=self._appcfg.suppress_logging,
            print_console=False
            )
        return self.api
    def __exit__(self, type, value, traceback):
        #Exception handling here
        loop=asyncio.get_event_loop()
        loop.create_task(self.api._session.close())
    def __del__(self):
        loop=asyncio.get_event_loop()
        loop.create_task(self.api._session.close())
    async def closesession(self):
         await self.api._session.close()
        
class MerakiApi:
    def __init__(self,appcfg):
        self._appcfg = appcfg
        self.api = meraki.DashboardAPI(
                api_key=self._appcfg.MERAKI_DASHBOARD_API_KEY,
                base_url=self._appcfg.meraki_base_url,
                simulate=self._appcfg.simulate,
                wait_on_rate_limit=self._appcfg.wait_on_rate_limit,
                log_file_prefix=os.path.basename(__file__)[:-3],
                log_path="Logs/",
                suppress_logging=self._appcfg.suppress_logging
            )
        logger = logging.getLogger('meraki')
        logger.setLevel(logging.WARNING)
    def getAPI(self):
        return self.api


import asyncio
from click import echo
from autosync import lib
import threading
from random import randrange
from meraki.exceptions import AsyncAPIError


async def sync(_db: object, _net_id: str, _net_name: str, _appcfg: object,
               _config: object, product: str,oFunc):
    """
    Proforms a full sync of the _config object that is passed to the fuction
    using the meraki dashboard SDK
    Args:
        oFunc: OverRide Functions
        _db(object): Meraki dashboard SDK Objeect
        _net_id(str): Current Network ID
        _net_name(str): Current Network Name
        _appcfg(object): Application Config
        _config(object): Current Network Configuration Object
        product(str):  Meraki Product being synced I.E. switch, wirless
    Returns:
	   Nothing Updates the _Config Object that is passed
    """
    if _appcfg.DEBUG:
        echo(
                f'Current Thread Name:{threading.currentThread().name} '
                f'Thread id:{threading.currentThread().native_id}')
    _maction = getattr(_db,product)
    settings = _config.settings()
    for attr in settings:


        waiting = randrange(0, 2)
        await asyncio.sleep(waiting)
        if _appcfg.DEBUG:
            echo(
                    f'\t {lib.bc.OKGREEN}Network:{_net_name}'
                    f'{lib.bc.OKBLUE} Requesting Config Object P{product} - {attr} '
                    f'in Orginization {threading.currentThread().name} with '
                    f'thread :{threading.currentThread().native_id} {lib.bc.Default}')
      
        if attr in oFunc:
            try:
                await eval(f'_config.Get_{attr}(_db, _net_id,_appcfg)')
            except AsyncAPIError as apie:
                echo(
                    f'\t {lib.bc.FAIL} Error Running Setting {attr} '
                    f'{lib.bc.WARNING}Error Message: {str(apie)}{lib.bc.Default}')
            except Exception as e:
                echo(
                        f'{lib.bc.FAIL}Network: {_net_name} '
                        f'{lib.bc.WARNING}Error with Module: {str(e)}'
                        f'{lib.bc.Default}'
                        f'Running OVerride Function {attr}')
        else:
            try:
                action = getattr(_maction, f'get{attr}')
                setattr(_config, attr, await action(_net_id))
            except AsyncAPIError as apie:
                echo(
                    f'\t {lib.bc.FAIL} Error Running Setting {attr} '
                    f'{lib.bc.WARNING}Error Message: {str(apie)}{lib.bc.Default}')
            except Exception as e:
                echo(
                        f'{lib.bc.FAIL}Network: {_net_name} '
                        f'{lib.bc.WARNING}Error with Module: {str(e)}'
                        f'{lib.bc.Default}')
                



from autosync import model


async def getOrginizationsAll(db: object, orgsdb):
	orgs = db.organizations.getOrganizations()
	for org in orgs:
		orgsdbtemp = model.ORGDB(org['id'], org['name'])
		orgsdb[org['id']] = orgsdbtemp


# @timeit.timer
async def getOrginizationsWhiteList(db, org, orgsdb):
	info = db.organizations.getOrganization(org)
	orgsdbtemp = model.ORGDB(info['id'], info['name'])
	orgsdb[info['id']] = orgsdbtemp



from autosync import lib
import copy
from autosync.mnetutils.ReSync import reSYNC
from click import echo
class GetAttr(type):
    def __getitem__(cls, x):
        return getattr(cls, x)

class networks(object):
    __metaclass__ = GetAttr
    
    # Dashboard Network Settings
    def __init__(self):
        self.Network = 'Test'
        self.NetworkAlertsSettings = None
        self.NetworkGroupPolicies = None
        self.NetworkTrafficAnalysis = None
        self.NetworkSyslogServers = None
        self.NetworkSnmp = None
        self.NetworkWebhooksHttpServers = None
        self.ReSync = False
    
    def settings(self):
        """
        
        Returns:
            Settings to cycle through when updating configurate
        """
        skip =  ['ReSync'] # Settings to skip
        items = [ i for i in self.__dict__.keys() if i not in skip]
        return items
    

    async def Update_NetworkSyslogServers(self, master: object,
                                          appcfg: object, db: object,
                                          net_id: str, task, netName):
        """
        Compaires Network Syslog Servers Configuration to Master and Updates
        Configuration if anything changes on target network
        Args:
            netName:
            task:
            master(Object): Current Master COnfiguration Object
            appcfg(Object): Applaciton configuration object
            db(object): Meraki SDK Object
            net_id(str): Network ID

        Returns:
            Nothing Updates meraki Dashbioard
        """
        ## Clone Syslog Settings
        if not await lib.compare(master.NetworkSyslogServers['servers'],
								 self.NetworkSyslogServers['servers']):
        
            # kinda works.... will trigger "change" if the 'roles' are unordered
            if appcfg.WRITE:
                echo(
                        f'\t{lib.bc.OKGREEN}Updating Syslog Settings in network {lib.bc.WARNING}{lib.bc.ENDC}')
                self.ReSync = True
                await db.updateNetworkSyslogServers(net_id,
                                                    **{'servers': []})
                await db.updateNetworkSyslogServers(net_id,
                                                    **master.NetworkSyslogServers)
        else:
            echo(
                    f'\t{lib.bc.OKBLUE} {netName} {lib.bc.OKGREEN}-Settings for  {task} Matched Master{lib.bc.ENDC}')
        if self.ReSync or not appcfg.CLEAN:
            await  reSYNC(db, net_id, self, 'NetworkSyslogServers'
                          , 'networks', appcfg)
            appcfg.CLEAN = True
            self.ReSync = True
        ## / end-Syslog Settings

    async def Update_NetworkWebhooksHttpServers(self,
                                                master: object,
                                                appcfg: object, db: object,
                                                net_id: str, task,
                                                netName):
        """
            Compaires Network WebHook Configuration to Master and Updates
            Configuration if anything changes on target network
            Args:
                netName:
                task:
                master(Object): Current Master COnfiguration Object
                appcfg(Object): Applaciton configuration object
                db(object): Meraki SDK Object
                net_id(str): Network ID

            Returns:
                Nothing Updates meraki Dashbioard
            """
        # Webhooks
        if not await lib.compare(master.NetworkWebhooksHttpServers,
								 self.NetworkWebhooksHttpServers):
            curr_list = []
            for cwh in self.NetworkWebhooksHttpServers:
                curr_list.append(cwh['name'])
            for mwh in master.NetworkWebhooksHttpServers:
                if not mwh['name'] in curr_list:
                    if appcfg.WRITE:
                        echo(
                            f'\t\t{lib.bc.OKBLUE}-Webhook {lib.bc.WARNING}{mwh["name"]}{lib.bc.ENDC}')
                        self.ReSync = True
                        mwh_tmp = copy.deepcopy(mwh)
                        mwh_tmp.pop('networkId')
                        await db.networks.createNetworkWebhooksHttpServer(
                                net_id, **mwh_tmp)
        else:
            echo(
                    f'\t{lib.bc.OKBLUE} {netName} {lib.bc.OKGREEN}-Settings for  {task} Matched Master{lib.bc.ENDC}')
        if self.ReSync or not appcfg.CLEAN:
            await  reSYNC(db, net_id, self,
                          'NetworkWebhooksHttpServers', 'networks', appcfg)
            appcfg.CLEAN = True
            self.ReSync = False

    async def Update_NetworkGroupPolicies(self,
                                          master: object,
                                          appcfg: object,
                                          db: object,
                                          net_id: str, task, netName):
        """
        Compaires Network Group Policy Configuration to Master and Updates
        Configuration if anything changes on target network
        Args:
            netName:
            task:
            master(Object): Current Master COnfiguration Object
            appcfg(Object): Applaciton configuration object
            db(object): Meraki SDK Object
            net_id(str): Network ID

        Returns:
            Nothing Updates meraki Dashbioard
                """
        # Group Policies
        # TODO Move this to Batch Action Could be faster
    
        for master_gp in master.NetworkGroupPolicies:
            tempGP = copy.deepcopy(master_gp)
            tempGP.pop('groupPolicyId')
            if appcfg.WRITE:
                local_gp = await lib.idFromName(
                    self.NetworkGroupPolicies,
                    tempGP['name'])
            
                if local_gp is None:
                    echo(f'\t\t{lib.bc.OKBLUE}Creating GP Policy named '
                          f'{tempGP["name"]}{lib.bc.ENDC}')
                    self.ReSync = True
                    try:
                        db.createNetworkGroupPolicy(net_id, **tempGP)
                    except:
                        echo(
                            f'{lib.bc.FAIL}ERROR: Cannot create GP policy named {tempGP["name"]}')
            
                else:
                    local_gpid = local_gp['groupPolicyId']
                    tempGP['groupPolicyId'] = local_gpid
                    if not await lib.soft_compare(tempGP,
												  await lib.idFromName(
                                                          master.NetworkGroupPolicies,
                                                          tempGP['name'])):
                        echo(
                            f'\t\t{lib.bc.OKBLUE}Updating GP Policy named {tempGP["name"]}{lib.bc.ENDC}')
                        self.ReSync = True
                        await db.networks.updateNetworkGroupPolicy(net_id,
                                                                   **tempGP)
        for localgp in self.NetworkGroupPolicies:
            mastergp = await lib.idFromName(self.NetworkGroupPolicies,
											localgp['name'])
            if mastergp is None:
                echo(
                    f't{lib.bc.WARNING}{netName} - {task}\t\t{lib.bc.OKBLUE}Removing GP Policy named {localgp["name"]}{lib.bc.ENDC} not found in Master')
                await db.deleteNetworkGroupPolicy(net_id, localgp['id'])
    
        if self.ReSync or not appcfg.CLEAN:
            await  reSYNC(db, net_id, self, 'NetworkGroupPolicies',
                          'networks', appcfg)
            appcfg.CLEAN = True
            self.ReSync = False





import sys
from os import getenv,getcwd,path,mkdir
from distutils.util import strtobool
from datetime import timedelta
from logging import getLevelName
import json

class APPCONFIG:
	def __init__(self,file=None):
		if not file is None:
			if str(file).endswith('.json'):
				self._openConfig(self._getdirectory(file))
			elif type(file) is dict:
				self._loadFromEnv(True)
				self._configFromFile(file)
			else:
				print('Incorrect File Format')
		else:
			self._loadFromEnv()
	
	
	def _openConfig(self,file):
		with open(file,'r') as f:
			_config =  json.loads(f.read())
			self._configFromFile(_config)

	def _configFromFile(self,config):
		for item in config:
			if config[item] is None:
				continue
			setattr(self,item,config[item])
		self.CACHE_DIR = self._getdirectory(self.CACHE_DIR)
		self.log_path = self._getdirectory(self.log_path)
		self.LOGGING_LEVEL = getLevelName(str(self.LOGGING_LEVEL).upper())
		if 'MERAKI_DASHBOARD_API_KEY' not in self.__dict__.keys() or self.MERAKI_DASHBOARD_API_KEY == 'None' \
				or self.MERAKI_DASHBOARD_API_KEY == "null":
			self.MERAKI_DASHBOARD_API_KEY = getenv('MERAKI_DASHBOARD_API_KEY', None)
			if self.MERAKI_DASHBOARD_API_KEY is None or self.MERAKI_DASHBOARD_API_KEY  == "null":
				print('Please add MERAKI_DASHBOARD_API_KEY by running "autosync config setkey <API Key>"'
				  	'or add MERAKI_DASHBOARD_API_KEY: <API Key > to your config.json file')
				sys.exit()
		
		
	def _loadFromEnv(self,override=False):
		self.USE_ENV = True
		self.MERAKI_DASHBOARD_API_KEY = getenv('MERAKI_DASHBOARD_API_KEY',None)
		if (self.MERAKI_DASHBOARD_API_KEY is None):
			if override:
				self.MERAKI_DASHBOARD_API_KEY = "null"
				pass
			else:
				print('Please Set MERAKI_DASHBOARD_API_KEY by running '
					  'autosync config setkey to contunue  or add it to the .env file ')
				sys.exit()
		self.meraki_base_url = getenv('meraki_base_url','https://api.meraki.com/api/v1/')
		self.simulate = bool(strtobool(getenv('simulate','False')))
		self.wait_on_rate_limit = bool(strtobool(getenv('wait_on_rate_limit','True')))
		self.maximum_concurrent_requests = (getenv('maximum_concurrent_requests',3))
		self.nginx_429_retry_wait_time = (getenv('nginx_429_retry_wait_time',8))
		self.maximum_retries = (getenv('maximum_retries',100))
		self.log_path = self._getdirectory(getenv('log_path','~/Logs'))
		self.suppress_logging = bool(strtobool(getenv('suppress_logging', 'False')))
		# Set this to FALSE for READ-ONLY, TRUE for "R/W"
		self.WRITE = bool(strtobool(getenv('WRITE','False')))
		# Set this to true, to crawl all networks. WARNING. Don't set WRITE
		# & ALL_ORGS unless you know what you're doing and dislike your job
		self.ALL_ORGS=bool(strtobool(getenv('ALL_ORGS','False')))
		if self.ALL_ORGS:
			self.whitelist = []
		else:
			## Only monitor these orgs, to keep the "crawl" down
			try:
				self.whitelist = list(getenv('whitelist').split(','))
			except:
				self.whitelist = []
		# Include switch settings?
		self.SWITCH = bool(strtobool(getenv('SWITCH','True')))
		# TARGET should be on ALL networks that are inscope, the master and
		# all the target networks
		self.tag_target = getenv('tag_target','autoSYNC')
		# MASTER should ONLY be on the 'golden network'
		self.tag_master = getenv('tag_master','master')
		#USed for Scale Testing Development
		self.tag_override = bool(strtobool(getenv('tag_override','False')))
		self.OpenRoaming = getenv('OpenRoaming','Meraki123')
		self.CACHE_DIR = self._getdirectory(str(getenv('CHACHE_DIR','~/mnetCache')))
		self.CACHE_TIMEOUT = int(getenv('CACHE_TIMEOUT','24'))
		self.USE_CACHE = bool(strtobool(getenv('USE_CACHE','True')))
		self.CLEAN = bool(strtobool(getenv('CLEAN','True')))
		self.RAD_KEYS_ALL= getenv('RAD_KEYS_ALL','Meraki123')
		try:
			self.SSID_SKIP_PSK=list(getenv('SSID_SKIP_PSK','SSID1,SSID2').split(','))
		except:
			self.SSID_SKIP_PSK = []
		self.LOGGING_LEVEL = getLevelName(str(getenv('LOGGING_LEVEL',"ERROR")).upper())
		self.TARGET_VLAN = bool(strtobool(getenv('TARGET_VLAN','True')))
		self.DEBUG = bool(strtobool(getenv('DEBUG','False')))

	
	def _getdirectory(self,location):
		if str(location).startswith('~/'):
			if not path.exists(path.expanduser(location)):
				if str(location).endswith('.json'):
					print('Config File Not Found')
					sys.exit()
				else:
					mkdir(path.expanduser(location))
			return path.expanduser(location)
		else:
			if not path.exists(path.abspath(location)):
				if str(location).endswith('.json'):
					print('Config File Not Found')
					sys.exit()
				else:
					mkdir(path.abspath(location))
			return path.abspath(location)

	def dumpConfigToFile(self):
		with open('config.json','w+') as f:
			c = self.__dict__
			print(json.dumps(c,indent=4,sort_keys=True))
			f.write(json.dumps(c, indent=4,sort_keys=True))
	
	def getRadSec(self):
		return getenv('RAD_KEYS_ALL')
	def checkCache(self):
		return timedelta(hours=self.CACHE_TIMEOUT)

		


import asyncio
import os
import click
from autosync import RUN


@click.group()
def cli():
    """ Tasks to start sync auto Sycn of Meraki Networks """
    pass


@click.command(help='Starts Meraki Dashboard Autosyn')
@click.option('-a', '--allOrgs')
@click.option('-c', '--useCache')
@click.option('-d','--debug')
@click.option('-f','--configfile')
@click.option('-k', '--merakiApiKey')
@click.option('-m', '--masterTag')
@click.option('-o', '--autoSyncOrgs')
@click.option('-s', '--suppressLogging')
@click.option('-t', '--targetTag')
@click.option('-T', '--cacheTimeOut')
@click.option('-w', '--write')
@click.option('--tagOverRide')
def start(suppresslogging=None, merakiapikey=None, write=None, allorgs=None, autosyncorgs=None,
          usecache=None, cachetimeout=None, mastertag=None, targettag=None, tagoverride=None,
          logginglevel=None,configfile=None,debug=None):
    if not configfile is None:
        asyncio.run(RUN(configfile))
    else:
        cfg ={}
        cfg['suppress_logging'] = suppresslogging
        cfg['MERAKI_DASHBOARD_API_KEY'] = merakiapikey
        cfg['WRITE'] = write
        cfg['whitelist'] = autosyncorgs
        cfg['tag_master'] = mastertag
        cfg['tag_target'] = targettag
        cfg['ALL_ORGS'] = allorgs
        cfg['USE_CACHE'] = usecache
        cfg['CACHE_TIMEOUT'] = cachetimeout
        cfg['tag_override'] = tagoverride
        cfg['LOGGING_LEVEL'] = logginglevel
        cfg['DEBUG'] = debug
        asyncio.run(RUN(cfg))




cli.add_command(start)



import requests, http3,json
async def _aironetie(net_id: str, _config: object, p_ssid: int,
                        p_apikey: str):
    # looks up org id for a specific org name
    # on failure returns 'null'
    try:
        client = http3.AsyncClient()
        r = await client.get(
            'https://api.meraki.com/api/v1/networks/%s/wireless/ssids/%s/overrides'
            % (net_id, p_ssid),
            headers={
                'X-Cisco-Meraki-API-Key': p_apikey,
                'Content-Type': 'application/json'
            })
    
        if r.status_code != requests.codes.ok:
            return 'null'
    
        rjson = r.json()
        return rjson
    except:
        return "null"

    
async def setAironetIe(net_id: str, cfg: object, _config: object, p_ssid: int,
                        p_data: dict):
    # looks up org id for a specific org name
    # on failure returns 'null'

    r = requests.put(
        'https://api.meraki.com/api/v1/networks/%s/wireless/ssids/%s/overrides'
        % (net_id, p_ssid),
        data=json.dumps(p_data),
        headers={
            'X-Cisco-Meraki-API-Key': cfg.MERAKI_DASHBOARD_API_KEY,
            'Content-Type': 'application/json'
        })

    if r.status_code != requests.codes.ok:
        return 'null'

    rjson = r.json()

    return rjson

async def AironetIe(net_id: str, _config: object, p_ssid: int,
                        p_apikey: str):
    # looks up org id for a specific org name
    # on failure returns 'null'
    client = http3.AsyncClient()
    r = await client.get(
        'https://api.meraki.com/api/v1/networks/%s/wireless/ssids/%s/overrides'
        % (net_id, p_ssid),
        headers={
            'X-Cisco-Meraki-API-Key': p_apikey,
            'Content-Type': 'application/json'
        })

    if r.status_code != requests.codes.ok:
        return 'null'

    rjson = r.json()

    return rjson

async def rfp_pwr(RFP):
    if 'twoFourGhzSettings' in RFP:
        if 'minPower' in RFP['twoFourGhzSettings'] and \
                RFP['twoFourGhzSettings']['minPower'] < 5:
            RFP['twoFourGhzSettings']['minPower'] = 5
        if 'maxPower' in RFP['twoFourGhzSettings'] and \
                RFP['twoFourGhzSettings']['maxPower'] < 5:
            RFP['twoFourGhzSettings']['maxPower'] = 5

    if 'fiveGhzSettings' in RFP:
        if 'minPower' in RFP['fiveGhzSettings'] and \
                RFP['fiveGhzSettings']['minPower'] < 5:
            RFP['fiveGhzSettings']['minPower'] = 8
        if 'maxPower' in RFP['fiveGhzSettings'] and \
                RFP['fiveGhzSettings']['maxPower'] < 5:
            RFP['fiveGhzSettings']['maxPower'] = 8
    return RFP



from autosync import lib
import copy
from meraki.exceptions import AsyncAPIError
from click import echo
class GetAttr(type):
    def __getitem__(cls, x):
        return getattr(cls, x)
class switch(object):
    def __init__(self):
        self.NetworkSwitchMtu = None
        self.NetworkSwitchSettings = 'Test'
        self.NetworkSwitchDscpToCosMappings = None
        self.NetworkSwitchRoutingMulticast = None
        self.NetworkSwitchAccessControlLists = None
        self.NetworkSwitchStormControl = None
        self.NetworkSwitchQosRules = None
        self.NetworkSwitchQosRulesOrder = None
        self.ReSync = False
    def settings(self):
        skip =  ['ReSync']
        items = [ i for i in self.__dict__.keys() if i not in skip]
        return items

    async def Update_NetworkSwitchAccessControlLists(self, master,
                                                     appcfg,
                                                     db, net_id, task,
                                                     netName):
        if not await lib.compare(master.NetworkSwitchAccessControlLists,
								 self.NetworkSwitchAccessControlLists):
            if appcfg.WRITE:
                acls = copy.deepcopy(
                        master.NetworkSwitchAccessControlLists)
                acls['rules'].remove(acls['rules'][len(acls[
                                                           'rules']) - 1])  # remove the default rule at the end
                self.ReSync = True
                echo(
                        f'\t {lib.bc.OKGREEN}-Updating Switch ACL rules...{lib.bc.ENDC}')
                await db.switch.updateNetworkSwitchAccessControlLists(
                        net_id, **acls)
        else:
            echo(
                    f'\t{lib.bc.OKBLUE} {netName} {lib.bc.OKGREEN}-Settings for  {task} Matched Master{lib.bc.ENDC}')

    async def Update_NetworkSwitchQosRules(self, master, appcfg, db,
                                           net_id, task, netName):
        if not await lib.soft_compare(master.NetworkSwitchQosRules,
									  self.NetworkSwitchQosRules):
        
            # {'ruleIds': ['577586652210270187', '577586652210270188', '577586652210270189']}
            rOrder_src = master.NetworkSwitchQosRules
            rOrder_dst = self.NetworkSwitchQosRules
            qosRuns = 0
            for rid in rOrder_src:
                for rid2 in rOrder_dst:
                    if rid['vlan'] is None or rid['vlan'] == rid2['vlan']:
                        if rid['protocol'] is None or rid['protocol'] == \
                                rid2['protocol']:
                            if rid['srcPort'] is None or rid['srcPort'] == \
                                    rid2['srcPort']:
                                if rid['dstPort'] is None or rid[
                                    'dstPort'] == rid2['dstPort']:
                                    if rid['dstPort'] is None or rid[
                                        'dscp'] == rid2['dscp']:
                                        continue
            
                ##	# echo(f'Duplicate rule, skipping!')
                #		continue
                if qosRuns == 0:
                    qosRuns += 1
                    echo(
                        f'\t{lib.bc.OKGREEN}-Cloning Switch QoS Rules...')
                # [{'id': '577586652210270187','vlan': None,'protocol': 'ANY','srcPort': None,'dstPort': None,'dscp': -1}, .. ]
                for r in master.NetworkSwitchQosRules:
                    if r['id'] == rid['id']:
                        rule = copy.deepcopy(r)
                        # try:
                        # pop the id, and srcPort/dstPort if they're empty, otherwismne it'll throw an error
                        rule.pop('id')
                        if rule['srcPort'] is None: rule.pop('srcPort')
                        if rule['dstPort'] is None: rule.pop('dstPort')
                        try:
                            if appcfg.WRITE:
                                await db.switch.reateNetworkSwitchQosRule(
                                        net_id, **rule)
                                echo(
                                        f'\t\t{lib.bc.OKGREEN}-Rule Created[{lib.bc.WARNING}{rule}{lib.bc.OKGREEN}]')
                        except AsyncAPIError as e:
                            echo(
                                f'\t {lib.bc.FAIL} Error running api: {lib.bc.WARNING} {str(e)}{lib.bc.Default}')
                    # except Exception as e:
                    #    echo(
                    #            f'\t\t{lib.bc.OKGREEN}-Rule already exists{lib.bc.ENDC} echo Error: {str(e)}')
                    else:
                        echo(
                            f'{lib.bc.FAIL}ERROR FINDING QoS RULE!!!{lib.bc.ENDC}')
    
        else:
            echo(
                    f'\t{lib.bc.OKBLUE} {netName} {lib.bc.OKGREEN}-Settings for  {task} Matched Master{lib.bc.ENDC}')




import threading
from autosync import lib
import asyncio
import logging
from .processor import validateNetworkProcessor
from click import echo 
class ValidateOrginization(threading.Thread):
	def __init__(self,org,appcfg,masternet):
		threading.Thread.__init__(self)
		self.org = org
		self.appcfg=appcfg
		self.oFunctions = None
		self.masternet = masternet
	def run(self):
		asyncio.run(self._asyncRun())
		lib.storeCache(self.org, self.appcfg)

	async def _asyncRun(self):
		echo(f'\tOrgName: {self.org.name}Thread PID:{threading.currentThread().native_id}')
		threading.currentThread().setName(self.org.name)
		echo(f'\tThread Name:{threading.currentThread().name}')
		with lib.MerakiAsyncApi(self.appcfg) as db:
			logger = logging.getLogger('meraki.aio')
			logger.setLevel(logging.WARNING)
			netCompareTask = [
					validateNetworkProcessor(self.org.networks[net], self.appcfg, self.masternet.networks['master'],db,self.oFunctions) for
					net in self.org.networks if self.appcfg.tag_master not in self.org.networks[net].tags]

			await asyncio.gather(*netCompareTask)


class ORGDB (object):
	def __init__(self,orgId,orgName):
		self.db = None
		self.id = orgId
		self.name = orgName
		self.syncruntime = None
		self.lastsync = None
		self.cached = False
		self.networks = {}
	def __dir__(self):
		return self.__dict__.keys()


# same as compare() but strips out ID/networkID for profiles/group policies etc
import copy
async def soft_compare(A, B):
	t_A = copy.deepcopy(A)
	t_B = copy.deepcopy(B)
	if 'id' in t_A: t_A.pop('id')
	if 'networkId' in t_A: t_A.pop('networkId')
	if 'groupPolicyId' in t_A: t_A.pop('groupPolicyId')
	if 'id' in t_B: t_B.pop('id')
	if 'networkId' in t_B: t_B.pop('networkId')
	if 'groupPolicyId' in t_B: t_B.pop('groupPolicyId')
	
	if 'dnsRewrite' in t_A: t_A.pop('dnsRewrite')
	if 'dnsRewrite' in t_B: t_B.pop('dnsRewrite')
	if 'adultContentFilteringEnabled' in t_A: t_A.pop(
		'adultContentFilteringEnabled')
	if 'adultContentFilteringEnabled' in t_B: t_B.pop(
		'adultContentFilteringEnabled')
	
	# had to add some logic to pop the "id" and "radsecEnabled". 'id' is unique and 'radsecEnabled' is beta for openroaming
	if 'radiusServers' in t_A:
		t_A['radiusServers'][0].pop('id')
		if 'radsecEnabled' in t_A['radiusServers'][0]:
			t_A['radiusServers'][0].pop('radsecEnabled')
		
		if 'radiusAccountingServers' in t_A:
			t_A['radiusAccountingServers'][0].pop('id')
			if 'radsecEnabled' in t_A['radiusAccountingServers'][0]:
				t_A['radiusAccountingServers'][0].pop('radsecEnabled')
	if 'radiusServers' in t_B:
		t_B['radiusServers'][0].pop('id')
		if 'radsecEnabled' in t_B['radiusServers'][0]:
			t_B['radiusServers'][0].pop('radsecEnabled')
		
		if 'radiusAccountingServers' in t_B:
			t_B['radiusAccountingServers'][0].pop('id')
			if 'radsecEnabled' in t_B['radiusAccountingServers'][0]:
				t_B['radiusAccountingServers'][0].pop('radsecEnabled')
	
	return await compare(t_A, t_B)


# compares JSON objects, directionarys and unordered lists will be equal
async def compare(A, B):
	result = True
	if A == None and B == None:
		return True
	if not type(A) == type(B):
		# print(f"Wrong type")
		return False
	try:
		if not type(A) == int and not type(A) == bool and not len(
				A) == len(B):
			# print(f'Not the same length')
			return False
	except:
		print()
	
	if type(A) == dict:
		for a in A:
			if a in B and not await compare(A[a], B[a]):
				return False
	elif type(A) == list:
		for a in A:
			if not a in B:
				return False
	else:
		if not A == B:
			return False
	return result
##END-OF COMPARE



import asyncio
import datetime
import random
from autosync import mnetutils
import time
import autosync.lib as lib
from click import echo

async def validateNetworkProcessor(networkobj: object,appcfg: object,masternet: object,db: object,oFunc: object):
    """
    ASYNC Function to synclye through configuration dict for network and call the clone functions based on netork id
    Args:

        oFunc:
        networkobj: Current Network Configuration
        masternet: Master Network Configuration
        appcfg: App Configuration Dict based on ENV or Configuration File
        db(object): Meraki SDKI Object
    Returns: Nothing

    """
    echo(f'{lib.bc.OKGREEN}Starting Network configuration validation...{lib.bc.ENDC}')
    for product in networkobj.dashboard:

        await mnetutils.clone(networkobj.dashboard[product], masternet.dashboard[product],
                              appcfg, db, networkobj.id, networkobj.name,
                              product, networkobj.functions[product]['update'])

           
           
   



async def proccessNetwork(net: object, appcfg: object, masternet: dict,
                          db):
    """

	Args:
        oFunc:
		net(object): Data Module contains configuration object for the network
		appcfg : App Configuration Dict based on ENV or Configuration File
		masternet: Diect of configuration for master object based on producted
		db (object): Meraki SDKI API Object

	Returns:

	"""
    t1 = time.perf_counter()
    echo(f'{lib.bc.OKGREEN}Started Configuration Sync at {t1:0.5f} secound'
          f' for Network: {net.name}')
    approvedList = net.supported
    if appcfg.tag_target in net.tags or appcfg.tag_override:
        if appcfg.tag_master in net.tags or net.id == 'L_575334852396597314':
            await mnetutils.sync(db, net.id, net.name, appcfg,
                                 masternet.dashboard['networks'], 'networks',
                                 masternet.functions['networks']['get'])
        else:
            await mnetutils.sync(db, net.id,net.name, appcfg,
                                 net.dashboard['networks'],'networks',net.functions['networks']['get'])
            
        for product in net.products:
            if product in approvedList:
                waiting = random.randrange(0, 5)
                await asyncio.sleep(waiting)
                if appcfg.tag_master in net.tags or net.id == 'L_575334852396597314':
                    await mnetutils.sync(db, net.id, net.name, appcfg,
                                         masternet.dashboard[product], product,
                                         masternet.functions[product]['get'])
                else:
                    await mnetutils.sync(db, net.id,net.name, appcfg,
                                         net.dashboard[product],product, net.functions[product]['get'])
            else:
                if  appcfg.DEBUG:
                    echo(f"No Sync Module for {product}")
                    
        net.lastsync = datetime.datetime.utcnow()
        net.syncruntime = time.perf_counter() - t1
        echo(f'{lib.bc.OKGREEN}Built Config From Network: {net.name} '
              f'Process took: {net.syncruntime:0.5f} secounds{lib.bc.Default}')
    
    
    else:
        t = time.perf_counter() - t1
        echo(f'Network: {net["name"]} '
              f'Skiped Syncing Process took: {t:0.5f} secounds')
        



import threading
import time
from autosync import lib
from autosync import model
import asyncio
import logging
from datetime import datetime
from .processor import proccessNetwork
from click import echo
class OrgSyncProcessor(threading.Thread):
    def __init__(self,org,appcfg,masternet,api):
        threading.Thread.__init__(self)
        self.org = org
        self.appcfg=appcfg
        self.masternet = masternet
        self.api = api
        
    def run(self):
        start_time = time.perf_counter()
        threading.currentThread().setName(self.org.name)
        if self.appcfg.DEBUG:
            echo(f'\tOrgName: {self.org.name} Sync Started at: {start_time} '
              f'Thread PID:{threading.currentThread().native_id}')
     
        nets = asyncio.run(self._getOrgNetworks())

        if self.appcfg.USE_CACHE:
            lib.loadCache(self.org, self.appcfg)
            self.org.name
        if not self.org.cached:
            for net in nets:
                if self.appcfg.tag_master in net['tags'] or net['id'] == 'L_575334852396597314':
                   self.masternet.networks['master'] = model.MNET(net)
                self.org.networks.update({net['id']:model.MNET(net)})
                #net_start_time = time.perf_counter()
                
                #self.org.networks[net['id']].syncruntime = \
                #    time.perf_counter() - net_start_time
            asyncio.run(self._asyncRun())
            self.org.syncruntime = time.perf_counter() - start_time
            self.org.lastsync = datetime.utcnow()
            lib.storeCache(self.org, self.appcfg)
        else:
            echo(f'OrgName: {self.org.name} Loaded From Cache Date: {self.org.lastsync}')


    async def _asyncRun(self):
        with lib.MerakiAsyncApi(self.appcfg) as db:
            logger = logging.getLogger('meraki.aio')
            logger.setLevel(logging.WARNING)
            netTasks = [proccessNetwork(self.org.networks[net], self.appcfg, self.masternet.networks['master'], db)
                        for net in self.org.networks]
            await asyncio.gather(*netTasks)
    
    async def _getOrgNetworks(self):
        with lib.MerakiAsyncApi(self.appcfg) as db:
            return  await db.organizations.getOrganizationNetworks(self.org.id)


from autosync import lib
from meraki.exceptions import AsyncAPIError
from .ReSync import reSYNC
from click import echo


async def clone(network, master, appcfg, db, net_id, netName, product,oFunc):
	merakiFunction = getattr(db,product)
	tasks = master.settings()
	for task in tasks:
		if task in oFunc:
			try:
				await eval(
					f"network.Update_{task}( master,appcfg,db,net_id,task,netName)")
			except Exception as e:
				echo(f'{lib.bc.FAIL} Network: {netName} ')
				echo(f'\t\t - {lib.bc.WARNING}Override for {product} Task: {task} failed with Error {e}{lib.bc.Default}')
		else:
			masterSetting = getattr(master, task)
			networkSetting = getattr(network, task)
			if not await lib.compare(masterSetting, networkSetting):
				echo(
					f'\t {lib.bc.OKGREEN}-Updating {task}...{lib.bc.ENDC}')
				if appcfg.WRITE:
					configure = getattr(merakiFunction, f"update{task}")
					try:
						await configure(net_id, **masterSetting)
						await  reSYNC(merakiFunction, net_id, network, task,
									  product,appcfg)
					except AsyncAPIError as e:
						echo(
							f'\t {lib.bc.FAIL} Error Running Setting {task} Error Message: {str(e)}  {lib.bc.Default}')
					appcfg.CLEAN = True
					network.ReSync = False
				else:
					echo(f"Write Disabled")
			else:
				echo(f'\t{lib.bc.OKBLUE} {netName} {lib.bc.OKGREEN}-Settings for  {task} Matched Master{lib.bc.ENDC}')



import copy
from autosync import lib
from click import echo


class GetAttr(type):
    def __getitem__(cls, x):
        return getattr(cls, x)


class wireless(object):
    __metaclass__ = GetAttr

    def __init__(self):
        # Dahsboard Wireless Settings
        self.NetworkWirelessSettings = None
        self.NetworkWirelessBluetoothSettings = None
        self.NetworkWirelessRfProfiles = None
        self.ssids_range = []  # should hold array of SSID_IDs, ex. [0,1,2,4,6,7]
        self.NetworkWirelessSsid = []
        self.NetworkWirelessSsidFirewallL3FirewallRules = []
        self.NetworkWirelessSsidFirewallL7FirewallRules = []
        self.NetworkWirelessSsidTrafficShapingRules = []
        self.NetworkWirelessSsidIdentityPsks = []
        self.ReSync = False
        self.hasAironetIE = None
        self.aironetie = None

    def settings(self):
        skip = ['ReSync', 'hasAironetIE', 'ssids_range']
        items = [i for i in self.__dict__.keys() if i not in skip]
        return items

    async def Get_NetworkWirelessSsid(self, db: object, net_id: str,
                                      _appcfg: object):
        ssids = []
        dashabordssids = await db.wireless.getNetworkWirelessSsids(net_id)
        for ssid in dashabordssids:
            ssids.append(ssid)
            if not "Unconfigured SSID" in ssid['name'] \
                    and not ssid['number'] in self.ssids_range:
                self.ssids_range.append(ssid['number'])
        self.NetworkWirelessSsid = ssids
        self.ssids_range

    async def Get_aironetie(self, db: object, net_id: str,
                            _appcfg: object):
        if self.hasAironetIE is None:
            # echo(f'Network {name} has aironetIE extensions!!!')
            self.hasAironetIE = False
        # echo(f'Network {name} needs aironetIE NFO')
        else:
            self.hasAironetIE = True

        # only do the full refresh if it's been cloned, cloneFrom_MR will set the aironetie = None
        if self.hasAironetIE:
            self.aironetie = []
            for i in range(0, 15):
                if i in self.ssids_range:  # only query/refresh the active SSIDS
                    aie_code = await lib._aironetie(net_id, self, i,
                                                    _appcfg.MERAKI_DASHBOARD_API_KEY)
                    echo(
                        f'\t\t\t{lib.bc.OKBLUE}Detecting AIE for SSID[{lib.bc.WARNING}{i}{lib.bc.OKBLUE}] '
                        f'Status[{lib.bc.WARNING}{aie_code}{lib.bc.OKBLUE}]{lib.bc.ENDC}')
                    self.aironetie.append(
                        aie_code)  # -1 for unkown, 0 for off, 1 for
            # self.aironetie
        # self.aironetie

    async def Get_NetworkWirelessSsidFirewallL3FirewallRules(self, db: object,
                                                             net_id: str,

                                                             _appcfg: object):
        ssids_l3 = []
        for ssid_num in range(0, 15):
            if ssid_num in self.ssids_range:
                ssids_l3.append(
                    await
                    db.wireless.getNetworkWirelessSsidFirewallL3FirewallRules(
                        net_id, ssid_num))
            else:
                ssids_l3.append([])
        self.NetworkWirelessSsidFirewallL3FirewallRules = ssids_l3

    async def Get_NetworkWirelessSsidFirewallL7FirewallRules(self, db: object,
                                                             net_id: str,
                                                             _appcfg: object):
        ssids_l7 = []
        for ssid_num in range(0, 15):
            if ssid_num in self.ssids_range:
                ssids_l7.append(
                    await
                    db.wireless.getNetworkWirelessSsidFirewallL7FirewallRules(
                        net_id, ssid_num))
            else:
                ssids_l7.append([])
        self.NetworkWirelessSsidFirewallL7FirewallRules = ssids_l7

    async def Get_NetworkWirelessSsidTrafficShapingRules(self, db: object,
                                                         net_id: str,

                                                         _appcfg: object):
        ssids_ts = []
        for ssid_num in range(0, 15):
            if ssid_num in self.ssids_range:
                ssids_ts.append(
                    await db.wireless.getNetworkWirelessSsidTrafficShapingRules(
                        net_id, ssid_num))
            else:
                ssids_ts.append([])
        self.NetworkWirelessSsidTrafficShapingRules = ssids_ts

    async def Get_NetworkWirelessSsidIdentityPsks(self, db: object, net_id: str,

                                                  _appcfg: object):
        NetworkWirelessSsidIdentityPsks = []
        for ssid_num in range(0, 15):
            if ssid_num in self.ssids_range:
                NetworkWirelessSsidIdentityPsks.append(await
                                                       db.wireless.getNetworkWirelessSsidIdentityPsks(
                                                           net_id,
                                                           ssid_num))
            else:
                NetworkWirelessSsidIdentityPsks.append([])
        self.NetworkWirelessSsidIdentityPsks = NetworkWirelessSsidIdentityPsks

    async def Update_NetworkWirelessSsid(self, master: object,
                                         appcfg: object, db: object,
                                         net_id: str, task, netName):
        """
        Compaires Wireless Network SSID Configuration to Master and Updates
        SSIDs if anything changes on target network
        Args:
            netName:
            task:
            
            master(Object): Current Master COnfiguration Object
            appcfg(Object): Applaciton configuration object
            db(object): Meraki SDK Object
            net_id(str): Network ID

        Returns:
            Nothing Updates meraki Dashbioard
        """
        # SSIDS
        # Process all SSIDs
        for i in range(0, 15):
            # Don't process SSIDs that are unconfigured
            if 'Unconfigured SSID' in self.NetworkWirelessSsid[i][
                'name'] \
                    and 'Unconfigured SSID' in \
                    master.NetworkWirelessSsid[i]['name']:
                continue

            if not await lib.soft_compare(master.NetworkWirelessSsid[i],
                                          self.NetworkWirelessSsid[i]):
                # Make a copy of the master SSID.... overrides will be needed to write
                temp_SSID = copy.deepcopy(master.NetworkWirelessSsid[i])
                echo(f'\t-{lib.bc.OKBLUE} SSID_Num[{i}] configuring '
                     f'SSID[{master.NetworkWirelessSsid[i]["name"]}] ')

                ###  START OF THE OVERRIDES/EXCEPTIONS
                if 'encryptionMode' in temp_SSID and temp_SSID[
                    'encryptionMode'] == 'wpa-eap':
                    temp_SSID['encryptionMode'] = 'wpa'
                # If the SSID has a single radius server,
                # it'll error if these are set to "None" so pop them
                if 'radiusFailoverPolicy' in temp_SSID and \
                        temp_SSID['radiusFailoverPolicy'] is None:
                    temp_SSID.pop('radiusFailoverPolicy')
                    # temp_SSID['radiusFailoverPolicy'] = 'Allow access'
                if 'radiusLoadBalancingPolicy' in temp_SSID and \
                        temp_SSID['radiusLoadBalancingPolicy'] is None:
                    temp_SSID.pop('radiusLoadBalancingPolicy')
                # this is to fix the case where the "target" network has
                # APvlanTags but the source does not. This wipes the target
                # if the source has no tags.
                if not 'apTagsAndVlanIds' in temp_SSID:
                    temp_SSID['apTagsAndVlanIds'] = []

                if 'radiusServers' in temp_SSID:
                    for rs in temp_SSID['radiusServers']:
                        rs['secret'] = appcfg['RAD_KEYS_ALL']

                if 'radiusAccountingServers' in temp_SSID:
                    for ras in temp_SSID['radiusAccountingServers']:
                        ras['secret'] = appcfg['RAD_KEYS_ALL']

                ### END OF THE OVERRIDES/EXCEPTIONS
                if appcfg.WRITE:
                    if 'name' in temp_SSID:
                        for ssid in appcfg.SSID_SKIP_PSK:
                            if self.NetworkWirelessSsid[i][
                                'name'] == ssid:
                                if 'psk' in temp_SSID:
                                    temp_SSID.pop('psk')
                    #               self.NetworkWirelessSsid[temp_SSID['number']] = \
                    await db.wireless.updateNetworkWirelessSsid(net_id,
                                                                **temp_SSID)
                    self.ReSync = True
                    if self.ReSync or not appcfg.CLEAN:
                        await self.Get_NetworkWirelessSsid(db, net_id,
                                                           self, appcfg)
                        appcfg.CLEAN = True
                        self.ReSync = False

            else:
                echo(
                    f'\t{lib.bc.OKBLUE} {netName} {lib.bc.OKGREEN}-Settings for  {task} Matched Master{lib.bc.ENDC}')

            # Clone the L3 FW rules

    async def Update_NetworkWirelessSsidFirewallL3FirewallRules(self,
                                                                master: object,
                                                                appcfg: object, db: object, net_id: str, task, netName):
        for ssid in self.NetworkWirelessSsid:
            i = ssid['number']
            if 'Unconfigured SSID' in self.NetworkWirelessSsid[i][
                'name'] or 'Unconfigured SSID' in \
                    master.NetworkWirelessSsid[i]['name']:
                pass
            else:
                if not await lib.compare(
                        self.NetworkWirelessSsidFirewallL3FirewallRules[
                            i],
                        master.NetworkWirelessSsidFirewallL3FirewallRules[
                            i]):
                    # echo(f'L3 is not the same')
                    echo(f'\t\t-{lib.bc.OKBLUE} Copied L3 rules for SSID['
                         f'{self.NetworkWirelessSsid[i]["name"]}] ')
                    lanAccess = True
                    l3rules = copy.deepcopy(
                        master.NetworkWirelessSsidFirewallL3FirewallRules[
                            i])
                    newL3 = {'rules': []}
                    for rule in l3rules['rules']:
                        if rule['destCidr'] == "Local LAN":
                            if rule['policy'] == "deny":
                                lanAccess = False
                            else:
                                lanAccess = True
                            # pull out the allow Lan Access rule, it's boolean
                            l3rules['rules'].remove(rule)
                        # pull out default rule, always the same
                        if rule['comment'] == "Default rule" \
                                or not rule['destCidr'] == "Local LAN":
                            newL3['rules'].append(rule)

                    # echo(f'L3 Rules are {newL3}')
                    newL3['allowLanAccess'] = lanAccess
                    if appcfg.WRITE:
                        self.NetworkWirelessSsidFirewallL3FirewallRules[
                            i] \
                            = await db.wireless.updateNetworkWirelessSsidFirewallL3FirewallRules(
                            net_id, i, **newL3)
                        self.ReSync = True
                        if self.ReSync or not appcfg.CLEAN:
                            await self.Get_NetworkWirelessSsidFirewallL3FirewallRules(
                                db, net_id, appcfg)
                            appcfg.CLEAN = True
                            self.ReSync = False
                else:
                    echo(
                        f'\t{lib.bc.OKBLUE} {netName} {lib.bc.OKGREEN}-Settings for  {task} Matched Master{lib.bc.ENDC}')

    async def Update_NetworkWirelessSsidFirewallL7FirewallRules(self,
                                                                master: object,
                                                                appcfg: object, db: object, net_id: str, task, netName):
        # Clone the L7 FW rules
        for ssid in self.NetworkWirelessSsid:
            i = ssid['number']
            if 'Unconfigured SSID' in self.NetworkWirelessSsid[i][
                'name'] \
                    or 'Unconfigured SSID' in \
                    master.NetworkWirelessSsid[i]['name']:
                continue
            if not await lib.compare(
                    self.NetworkWirelessSsidFirewallL7FirewallRules[i],
                    master.NetworkWirelessSsidFirewallL7FirewallRules[i]):
                l7rules = \
                    copy.deepcopy(
                        master.NetworkWirelessSsidFirewallL7FirewallRules[
                            i])
                # echo(f'L7 not the same ... cloning')
                echo(f'\t\t-{lib.bc.OKBLUE} '
                     f'Copied L7 rules for '
                     f'SSID[{self.NetworkWirelessSsid[i]["name"]}] ')

                if appcfg.WRITE:
                    self.NetworkWirelessSsidFirewallL7FirewallRules[i] = \
                        await db.wireless.updateNetworkWirelessSsidFirewallL7FirewallRules(
                            net_id, i, **l7rules)
                    self.ReSync = True
                    if self.ReSync or not appcfg.CLEAN:
                        await self.Get_NetworkWirelessSsidFirewallL7FirewallRules(
                            db, net_id, appcfg)
                        appcfg.CLEAN = True
                        self.ReSync = False
            else:
                echo(
                    f'\t{lib.bc.OKBLUE} {netName} {lib.bc.OKGREEN}-Settings for  {task} Matched Master{lib.bc.ENDC}')
        # Clone the TS Rules

    async def Update_NetworkWirelessSsidTrafficShapingRules(self,
                                                            master: object,
                                                            appcfg: object, db: object, net_id: str, task, netName):
        for ssid in self.NetworkWirelessSsid:
            i = ssid['number']
            if 'Unconfigured SSID' in self.NetworkWirelessSsid[i][
                'name'] \
                    or 'Unconfigured SSID' in \
                    master.NetworkWirelessSsid[i]['name']:
                continue
            if not await lib.compare(
                    self.NetworkWirelessSsidTrafficShapingRules[i],
                    master.NetworkWirelessSsidTrafficShapingRules[i]):
                echo(f'\t\t-{lib.bc.OKBLUE} Copied Traffic '
                     f'Shaping rules for SSID[{self.NetworkWirelessSsid[i]["name"]}] ')
                try:
                    TSrules = copy.deepcopy(
                        master.NetworkWirelessSsidTrafficShapingRules[
                            i])
                    if appcfg.WRITE:
                        self.NetworkWirelessSsidTrafficShapingRules[i] \
                            = await db.wireless.updateNetworkWirelessSsidTrafficShapingRules(
                            net_id, i, **TSrules)
                        self.ReSync = True

                        await self.Get_NetworkWirelessSsidTrafficShapingRules(
                            db, net_id, self, appcfg)
                        appcfg.CLEAN = True
                        self.ReSync = False
                except Exception as e:
                    echo(
                        f'\t\t-{lib.bc.FAIL}Failed to update TrafficShaping.'
                        f' Make sure all rules are complete{lib.bc.ENDC} '
                        f'Error Code: {str(e)}')
            else:
                echo(
                    f'\t{lib.bc.OKBLUE} {netName} {lib.bc.OKGREEN}-Settings for  {task} Matched Master{lib.bc.ENDC}')
        # this also updates ssids_range

    async def Update_aironetie(self, master: object,
                               appcfg: object, db: object, net_id: str,
                               task, netName):
        for i in self.ssids_range:  # and self.hasAironetIE:
            if 'Unconfigured SSID' in self.NetworkWirelessSsid[i][
                'name'] \
                    or 'Unconfigured SSID' in \
                    master.NetworkWirelessSsid[i]['name']:
                continue
            if self.hasAironetIE and not await lib.compare(
                    self.aironetie[i],
                    master.aironetie[i]):
                if appcfg.WRITE:
                    self.ReSync = True
                    self.setaironetie(net_id, i, master.aironetie[i])
                    echo(
                        f'{lib.bc.OKBLUE}\t\tConfiguring AironetIE[{lib.bc.WARNING}'
                        f'{master.aironetie[i]}'
                        f'{lib.bc.OKBLUE}] on SSID[{lib.bc.WARNING}'
                        f'{i}{lib.bc.OKBLUE}]{lib.bc.ENDC}')
                    if self.hasAironetIE:
                        await self.Get_aironetie(net_id, appcfg)
            else:
                echo(
                    f'\t{lib.bc.OKBLUE} {netName} {lib.bc.OKGREEN}-Settings for  {task} Matched Master{lib.bc.ENDC}')

    async def Update_NetworkWirelessRfProfiles(self,
                                               master: object,
                                               appcfg: object, db: object,
                                               net_id: str, task, netName):
        """
        Compaires Wireless Network RF Profiles with the Master Configuratio
        and Updates The RF Profile If neede
        Args:
            netName:
            task:
            master(Object): Current Master COnfiguration Object
            appcfg(Object): Applaciton configuration object
            db(object): Meraki SDK Object
            net_id(str): Network ID

        Returns:
            Nothing Updates meraki Dashbioard
        """
        # RFProfiles - (if it exists and not equal, delete/update. If it doesn't exist, create)
        self_RFPS = copy.deepcopy(self.NetworkWirelessRfProfiles)
        master_RFPS = copy.deepcopy(master.NetworkWirelessRfProfiles)
        if not self_RFPS is None:
            for srfp in self_RFPS:
                srfp.pop('id')
                srfp.pop('networkId')
        if not master_RFPS is None:
            for mrfp in master_RFPS:
                mrfp.pop('id')
                mrfp.pop('networkId')

        if not await lib.compare(self_RFPS,
                                 master_RFPS):  # Profiles are NOT the same
            for masterRF in master.NetworkWirelessRfProfiles:
                for selfRF in self.NetworkWirelessRfProfiles:
                    if masterRF['name'] == selfRF['name']:
                        # echo(f'RF Profile[{masterRF["name"]}] FOUND')
                        if not await lib.soft_compare(
                                masterRF,
                                selfRF):  # It's in there but might not be the same
                            echo(
                                f'\t{lib.bc.OKBLUE}RF Profile[{lib.bc.WARNING}{masterRF["name"]}{lib.bc.OKBLUE}] !!! Updating RF Profile{lib.bc.ENDC}'
                            )
                            newRF = copy.deepcopy(masterRF)
                            newRF.pop('id')
                            newRF.pop('networkId')
                            newRF.pop('name')
                            newRF['fiveGhzSettings'].pop(
                                'validAutoChannels')
                            newRF['fiveGhzSettings'][
                                'validAutoChannels'] = [
                                36, 40, 44, 48, 52, 56, 60, 64, 100,
                                104,
                                108, 112,
                                116, 120, 124, 128, 132, 136, 140, 144,
                                149, 153,
                                157, 161, 165
                            ]
                            newRF = await lib.rfp_pwr(newRF)
                            if appcfg.WRITE:
                                await db.wireless.updateNetworkWirelessRfProfile(
                                    net_id, selfRF['id'], **newRF)
                                self.ReSync = True
                        else:
                            # no more RFProfiles in self, create one
                            echo(
                                f'\t{lib.bc.OKBLUE}RF Profile[{lib.bc.WARNING}{masterRF["name"]}{lib.bc.OKBLUE}]!!! New RFP created in network{lib.bc.ENDC}'
                            )
                            newRF = copy.deepcopy(masterRF)
                            newRF.pop('id')
                            newRF.pop('networkId')
                            newRF = await lib.rfp_pwr(newRF)
                            newRF['fiveGhzSettings'].pop(
                                'validAutoChannels')
                            newRF['fiveGhzSettings'][
                                'validAutoChannels'] = [
                                36, 40, 44, 48, 52, 56, 60, 64, 100,
                                104,
                                108, 112, 116,
                                120, 124, 128, 132, 136, 140, 144, 149,
                                153, 157, 161, 165
                            ]
                            if appcfg.WRITE:
                                await db.createNetworkWirelessRfProfile(
                                    net_id, **newRF)
                                self.ReSync = True
        else:
            echo(
                f'\t{lib.bc.OKBLUE} {netName} {lib.bc.OKGREEN}-Settings for  {task} Matched Master{lib.bc.ENDC}')

    async def Update_NetworkWirelessSsidIdentityPsks(self,
                                                     master: object,
                                                     appcfg: object,
                                                     db: object,
                                                     net_id: str, task,
                                                     netName):
        """
        Compaires SSID iPSKs with the Master Configuratio
        and Updates The RF Profile If neede
        Args:
            netName:
            task:

            master(Object): Current Master COnfiguration Object
            appcfg(Object): Applaciton configuration object
            db(object): Meraki SDK Object
            net_id(str): Network ID

        Returns:
            Nothing Updates meraki Dashbioard
        """
        # NetworkWirelessSsidIdentityPsks

        ipsk_tmp = []
        for r in range(0, 15):
            ipsk_tmp.append({})
        for ssid_num in self.ssids_range:
            # if not ssid_num in self.ssids_range: continue
            # ipsk_tmp.append({}) #keep track of master iPSKs so we can remove unused ones from local(self)
            for m_ipsk in master.NetworkWirelessSsidIdentityPsks[ssid_num]:
                if not m_ipsk['name'] in ipsk_tmp[ssid_num]:
                    ipsk_tmp[ssid_num][m_ipsk['name']] = m_ipsk[
                        'passphrase']

                # ipsks are not empty, find the matching group policy
                new_ipsk = copy.deepcopy(m_ipsk)
                # pop off the ID from master, new one will be created "local"
                new_ipsk.pop('id')
                master_GP_tmp = await lib.matchGidByName(
                    master.NetworkGroupPolicies,
                    str(new_ipsk['groupPolicyId']))
                local_GP_tmp = await lib.idFromName(
                    self.NetworkGroupPolicies,
                    str(master_GP_tmp['name']))
                new_ipsk['groupPolicyId'] = local_GP_tmp['groupPolicyId']

                for s_ipsk in self.NetworkWirelessSsidIdentityPsks[
                    ssid_num]:
                    if new_ipsk['name'] == s_ipsk['name']:
                        # if passwords are different, delete the ipsk and re-create
                        if new_ipsk['passphrase'] != s_ipsk['passphrase']:
                            if appcfg.WRITE:
                                self.ReSync = True
                                try:
                                    await db.wireless.deleteNetworkWirelessSsidIdentityPsk(
                                        net_id, ssid_num, s_ipsk['id'])
                                except:
                                    echo(
                                        f'ERROR: iPSK Issue, resyncing and trying again')
                                    await self.Get_NetworkWirelessSsidIdentityPsks(
                                        db, net_id, appcfg)
                                    await db.wireless.deleteNetworkWirelessSsidIdentityPsk(
                                        net_id, ssid_num, s_ipsk['id'])

                    else:
                        try:
                            await db.wireless.createNetworkWirelessSsidIdentityPsk(
                                net_id, ssid_num, **new_ipsk)
                            self.ReSync = True
                        except:
                            echo(
                                f'{lib.bc.FAIL}{netName} - {task} \t\t{lib.bc.FAIL}iPSK already created or still there{lib.bc.ENDC}')

        if self.ReSync or not appcfg.CLEAN:
            await self.Get_NetworkWirelessSsidIdentityPsks(db, net_id)
            appcfg.CLEAN = True
            self.ReSync = False

        # cleanUP local iPSK
        for ssid_num in self.ssids_range:
            for s_ipsk in self.NetworkWirelessSsidIdentityPsks[
                ssid_num]:
                if not s_ipsk['name'] in ipsk_tmp[ssid_num]:
                    if appcfg.WRITE:
                        self.ReSync = True
                        echo(
                            f'\t\t{lib.bc.OKBLUE}-Removing Legacy iPSK[{s_ipsk["name"]}]{lib.bc.ENDC}')
                        await db.wireless.deleteNetworkWirelessSsidIdentityPsk(
                            net_id, ssid_num, s_ipsk['id'])



from .find import idFromName,matchGidByName
from .helpers import _aironetie,setAironetIe,rfp_pwr
from .bcolors import bcolors as bc
from .merakiapi import MerakiApi,MerakiAsyncApi
from .netcompare import compare,soft_compare
from .cache import storeCache,loadCache,clearCache